"""Allows useful indirection to test Pin naming logic by switching platform in testing
    or provide bootstrapping logic for board identification where auto-detection is not
    feasible (e.g. multiple ESP8266 boards architecturally identical). Once runtime
    environment is established, can choose various routes to make available and re-export
    common modules and operations, depending on platform support
"""
import sys

# We intentionally are patching into this namespace as module names so skip the name check.
# pylint: disable=invalid-name
platform = sys.platform

board_id = None

if platform is not None:
    if platform == "esp8266":  # TODO more conservative board-guessing
        board_id = "feather_huzzah"
    elif platform == "samd21":
        board_id = "feather_m0_express"
    elif platform == "pyboard":
        platform = "stm32"
        board_id = "pyboard"
    elif platform == "linux":
        from Adafruit_GPIO import Platform
        if Platform.platform_detect() == Platform.RASPBERRY_PI:
            if Platform.pi_version() == 1:
                board_id = "raspi_1"
            elif Platform.pi_version() == 2:
                board_id = "raspi_2"
            elif Platform.pi_version() == 3:
                board_id = "raspi_3"
        elif Platform.platform_detect() == Platform.BEAGLEBONE_BLACK:
            board_id = "beaglebone_black"

implementation = sys.implementation.name
if implementation == "micropython":
    from utime import sleep
elif implementation == "circuitpython" or implementation == "cpython":
    from time import sleep



import Adafruit_BBIO.GPIO as GPIO

# Pins dont exist in CPython so...lets make our own!
class Pin:
    IN = 0
    OUT = 1
    LOW = 0
    HIGH = 1
    PULL_NONE = 0
    PULL_UP = 1
    PULL_DOWN = 2
    
    id = None
    _value = LOW
    _mode = IN
    
    def __init__(self, pin_name):
        self.id = pin_name

    def __repr__(self):
        return str(self.id)

    def __eq__(self, other):
        return self.id == other

    def init(self, mode=IN, pull=None):
        if mode != None:
            if mode == self.IN:
                self._mode = self.IN
                GPIO.setup(self.id, GPIO.IN)
            elif mode == self.OUT:
                self._mode = self.OUT
                GPIO.setup(self.id, GPIO.OUT)
            else:
                raise RuntimeError("Invalid mode for pin: %s" % self.id)
        if pull != None:
            if self._mode != self.IN:
                raise RuntimeError("Cannot set pull resistor on output")
            if pull == self.PULL_UP:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            elif pull == self.PULL_DOWN:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
            else:
                raise RuntimeError("Invalid pull for pin: %s" % self.id)       

    def value(self, val=None):
        if val != None:
            if val == self.LOW:
                self._value = val
                GPIO.output(self.id, val)
            elif val == self.HIGH:
                self._value = val
                GPIO.output(self.id, val)
            else:
                raise RuntimeError("Invalid value for pin")
        else:
            return GPIO.input(self.id)

P8_3 = Pin('P8_3')
P8_4 = Pin('P8_4')
P8_5 = Pin('P8_5')
P8_6 = Pin('P8_6')
P8_7 = Pin('P8_7')
P8_8 = Pin('P8_8')
P8_9 = Pin('P8_9')
P8_10 = Pin('P8_10')
P8_11 = Pin('P8_11')
P8_12 = Pin('P8_12')
P8_13 = Pin('P8_13')
P8_14 = Pin('P8_14')
P8_15 = Pin('P8_15')
P8_16 = Pin('P8_16')
P8_17 = Pin('P8_17')
P8_18 = Pin('P8_18')
P8_19 = Pin('P8_19')
P8_20 = Pin('P8_20')
P8_21 = Pin('P8_21')
P8_22 = Pin('P8_22')
P8_23 = Pin('P8_23')
P8_24 = Pin('P8_24')
P8_25 = Pin('P8_25')
P8_26 = Pin('P8_26')
P8_27 = Pin('P8_27')
P8_28 = Pin('P8_28')
P8_29 = Pin('P8_29')
P8_30 = Pin('P8_30')
P8_31 = Pin('P8_31')
P8_32 = Pin('P8_32')
P8_33 = Pin('P8_33')
P8_34 = Pin('P8_34')
P8_35 = Pin('P8_35')
P8_36 = Pin('P8_36')
P8_37 = Pin('P8_37')
P8_38 = Pin('P8_38')
P8_39 = Pin('P8_39')
P8_40 = Pin('P8_40')
P8_41 = Pin('P8_41')
P8_42 = Pin('P8_42')
P8_43 = Pin('P8_43')
P8_44 = Pin('P8_44')
P8_45 = Pin('P8_45')
P8_46 = Pin('P8_46')
P9_11 = Pin('P9_11')
P9_12 = Pin('P9_12')
P9_13 = Pin('P9_13')
P9_14 = Pin('P9_14')
P9_15 = Pin('P9_15')
P9_16 = Pin('P9_16')
P9_17 = Pin('P9_17')
P9_18 = Pin('P9_18')
P9_19 = Pin('P9_19')
P9_20 = Pin('P9_20')
P9_21 = Pin('P9_21')
P9_22 = Pin('P9_22')
P9_23 = Pin('P9_23')
P9_24 = Pin('P9_24')
P9_25 = Pin('P9_25')
P9_26 = Pin('P9_26')
P9_27 = Pin('P9_27')
P9_28 = Pin('P9_28')
P9_29 = Pin('P9_29')
P9_30 = Pin('P9_30')
P9_31 = Pin('P9_31')
P9_41 = Pin('P9_41')
P9_42 = Pin('P9_42')

USR0 = Pin('USR0')
USR1 = Pin('USR1')
USR2 = Pin('USR2')
USR3 = Pin('USR3')

SCL = Pin('P9_19')
SDA = Pin('P9_20')

# Refer to header default pin modes
# http://beagleboard.org/static/images/cape-headers.png
#
# P9_17 (SPI0_CSO => CE0) enables peripheral device
# P9_18 (SPI0_D1 => MOSI) outputs data to peripheral device
# P9_21 (SPIO_DO => MISO) receives data from peripheral device
# P9_22 (SPI0_SCLK => SCLK) outputs clock signal
# 
# Use config-pin to set pin mode for SPI pins
# https://github.com/beagleboard/bb.org-overlays/tree/master/tools/beaglebone-universal-io
# config-pin p9.17 spi_cs
# config-pin p9.18 spi
# config-pin p9.21 spi
# config-pin p9.22 spi_sclk
#
CE0 = Pin('P9_17')
MOSI = Pin('P9_18')
MISO = Pin('P9_21')
SCLK = Pin('P9_22')
#CircuitPython naming convention for SPI Clock
SCK = Pin('P9_22')

# Pins for SPI1
# refer to:
# http://beagleboard.org/static/images/cape-headers-spi.png
#
# CE1 P9.28 SPI1_CS0
# MISO_1 P9.29 SPI1_D0
# MOSI_1 P9.30 SPI1_D1
# SCLK_1 P9.31 SPI_SCLK
#
# SPI1 conflicts with HDMI Audio (McASP)
#
# Refer to:
# https://elinux.org/Beagleboard:BeagleBoneBlack_Debian#U-Boot_Overlays
#
# To Disable HDMI AUDIO, uncomment this line in /boot/uEnv.txt:
# disable_uboot_overlay_audio=1
#
# Set pin modes for SPI1 with:
#
# config-pin p9.28 spi1_cs
# config-pin p9.29 spi1
# config-pin p9.30 spi1
# config-pin p9.31 spi_sclk
CE1 = Pin('P9_28')
MOSI_1 = Pin('P9_29')
MISO_1 = Pin('P9_30')
SCLK_1 = Pin('P9_31')
#CircuitPython naming convention for SPI Clock
SCK_1 = Pin('P9_31')

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((0, SCLK, MOSI, MISO), (1, SCLK_1, MOSI_1, MISO_1))

# ordered as uartId, txId, rxId
uartPorts = (
    (),
)

i2cPorts = (
    (2, SCL, SDA),
)




from adafruit_blinka.microcontroller.beaglebone_black import pin

# initial pins, to mimic bonescript demo
P8_3 = pin.P8_3
P8_4 = pin.P8_4
P8_5 = pin.P8_5
P8_6 = pin.P8_6
P8_7 = pin.P8_7
P8_8 = pin.P8_8
P8_9 = pin.P8_9
P8_10 = pin.P8_10
P8_11 = pin.P8_11
P8_12 = pin.P8_12
P8_13 = pin.P8_13
P8_14 = pin.P8_14
P8_15 = pin.P8_15
P8_16 = pin.P8_16
P8_17 = pin.P8_17
P8_18 = pin.P8_18
P8_19 = pin.P8_19
P8_20 = pin.P8_20
P8_21 = pin.P8_21
P8_22 = pin.P8_22
P8_23 = pin.P8_23
P8_24 = pin.P8_24
P8_25 = pin.P8_25
P8_26 = pin.P8_26
P8_27 = pin.P8_27
P8_28 = pin.P8_28
P8_29 = pin.P8_29
P8_30 = pin.P8_30
P8_31 = pin.P8_31
P8_32 = pin.P8_32
P8_33 = pin.P8_33
P8_34 = pin.P8_34
P8_35 = pin.P8_35
P8_36 = pin.P8_36
P8_37 = pin.P8_37
P8_38 = pin.P8_38
P8_39 = pin.P8_39
P8_40 = pin.P8_40
P8_41 = pin.P8_41
P8_42 = pin.P8_42
P8_43 = pin.P8_43
P8_44 = pin.P8_44
P8_45 = pin.P8_45
P8_46 = pin.P8_46
P9_11 = pin.P9_11
P9_12 = pin.P9_12
P9_13 = pin.P9_13
P9_14 = pin.P9_14
P9_15 = pin.P9_15
P9_16 = pin.P9_16
P9_17 = pin.P9_17
P9_18 = pin.P9_18
P9_19 = pin.P9_19
P9_20 = pin.P9_20
P9_21 = pin.P9_21
P9_22 = pin.P9_22
P9_23 = pin.P9_23
P9_24 = pin.P9_24
P9_25 = pin.P9_25
P9_26 = pin.P9_26
P9_27 = pin.P9_27
P9_28 = pin.P9_28
P9_29 = pin.P9_29
P9_30 = pin.P9_30
P9_31 = pin.P9_31
P9_41 = pin.P9_41
P9_42 = pin.P9_42

LED_USR0 = pin.USR0
LED_USR1 = pin.USR1
LED_USR2 = pin.USR2
LED_USR3 = pin.USR3

# I2C and SPI pins from:
# src/adafruit_blinka/board/raspi_23.py
# SDA = pin.SDA
# SCL = pin.SCL
# CE1 = pin.D7
# CE0 = pin.D8
# MISO = pin.D9
# MOSI = pin.D10
# SCLK = pin.D11
# SCK = pin.D11
# TXD = pin.D14
# RXD = pin.D15
# MISO_1 = pin.D19
# MOSI_1 = pin.D20
# SCLK_1 = pin.D21
# SCK_1 = pin.D21

SDA = pin.SDA
SCL = pin.SCL

# Refer to header default pin modes
# http://beagleboard.org/static/images/cape-headers.png
# P9_17 (SPI0_CSO => CE0) enables peripheral device
# P9_18 (SPI0_D1 => MOSI) outputs data to peripheral device
# P9_21 (SPIO_DO => MISO) receives data from peripheral device
# P9_22 (SPI0_SCLK => SCLK) outputs clock signal
CE0 = pin.CE0
MOSI = pin.MOSI
MISO = pin.MISO
SCLK = pin.SCLK
#CircuitPython naming convention for SPI Clock
SCK = pin.SCK

# SPI1 pins
# http://beagleboard.org/static/images/cape-headers-spi.png
CE1 = pin.CE1
MOSI_1 = pin.MOSI_1
MISO_1 = pin.MISO_1
SCLK_1 = pin.SCLK_1
#CircuitPython naming convention for SPI Clock
SCK_1 = pin.SCK_1



import time
import math
import _rpi_ws281x as ws
import atexit

# LED configuration.
LED_CHANNEL    = 0
LED_FREQ_HZ    = 800000     # Frequency of the LED signal.  We only support 800KHz
LED_DMA_NUM    = 10         # DMA channel to use, can be 0-14.
LED_BRIGHTNESS = 255        # We manage the brightness in the neopixel library
LED_INVERT     = 0          # We don't support inverted logic
LED_STRIP      = None # We manage the color order within the neopixel library

# a 'static' object that we will use to manage our PWM DMA channel
# we only support one LED strip per raspi
_led_strip = None

def neopixel_write(gpio, buf):
    global _led_strip # we'll have one strip we init if its not at first

    if _led_strip is None:
        # Create a ws2811_t structure from the LED configuration.
        # Note that this structure will be created on the heap so you
        # need to be careful that you delete its memory by calling
        # delete_ws2811_t when it's not needed.
        _led_strip = ws.new_ws2811_t()

        # Initialize all channels to off
        for channum in range(2):
            channel = ws.ws2811_channel_get(_led_strip, channum)
            ws.ws2811_channel_t_count_set(channel, 0)
            ws.ws2811_channel_t_gpionum_set(channel, 0)
            ws.ws2811_channel_t_invert_set(channel, 0)
            ws.ws2811_channel_t_brightness_set(channel, 0)

        channel = ws.ws2811_channel_get(_led_strip, LED_CHANNEL)

        # Initialize the channel in use
        count = 0
        if len(buf) % 3 == 0:            
            # most common, divisible by 3 is likely RGB
            LED_STRIP = ws.WS2811_STRIP_RGB
            count = len(buf)//3
        elif len(buf) % 4 == 0:
            LED_STRIP = ws.SK6812_STRIP_RGBW
            count = len(buf)//4
        else:
            raise RuntimeError("We only support 3 or 4 bytes-per-pixel")

        ws.ws2811_channel_t_count_set(channel, count) # we manage 4 vs 3 bytes in the library
        ws.ws2811_channel_t_gpionum_set(channel, gpio._pin.id)
        ws.ws2811_channel_t_invert_set(channel, LED_INVERT)
        ws.ws2811_channel_t_brightness_set(channel, LED_BRIGHTNESS)
        ws.ws2811_channel_t_strip_type_set(channel, LED_STRIP)

        # Initialize the controller
        ws.ws2811_t_freq_set(_led_strip, LED_FREQ_HZ)
        ws.ws2811_t_dmanum_set(_led_strip, LED_DMA_NUM)
    
        resp = ws.ws2811_init(_led_strip)
        if resp != ws.WS2811_SUCCESS:
            message = ws.ws2811_get_return_t_str(resp)
            raise RuntimeError('ws2811_init failed with code {0} ({1})'.format(resp, message))
        atexit.register(neopixel_cleanup)

    channel = ws.ws2811_channel_get(_led_strip, LED_CHANNEL)
    if gpio._pin.id != ws.ws2811_channel_t_gpionum_get(channel):
        raise RuntimeError("Raspberry Pi neopixel support is for one strip only!")

    # assign all colors!
    for i in range(len(buf) // 3):
        r = buf[3*i]
        g = buf[3*i+1]
        b = buf[3*i+2]
        pixel = (r << 16) | (g << 8) | b
        ws.ws2811_led_set(channel, i, pixel)
    
    resp = ws.ws2811_render(_led_strip)
    if resp != ws.WS2811_SUCCESS:
        message = ws.ws2811_get_return_t_str(resp)
        raise RuntimeError('ws2811_render failed with code {0} ({1})'.format(resp, message))
    time.sleep(0.001 * ((len(buf)//100)+1))  # about 1ms per 100 bytes


def neopixel_cleanup():
    global _led_strip

    if _led_strip is not None:
        # Ensure ws2811_fini is called before the program quits.
        ws.ws2811_fini(_led_strip)
        # Example of calling delete function to clean up structure memory.  Isn't
        # strictly necessary at the end of the program execution here, but is good practice.
        ws.delete_ws2811_t(_led_strip)
        _led_strip = None
